# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
import warnings

import pandas as pd

from chronix2grid import constants
from chronix2grid import utils
from chronix2grid.generation import generation_utils

from chronix2grid.generation.dispatch import EconomicDispatch


# MSG_PYPSA_DEPENDENCY = "Please install PypsaDispatchBackend dependency to launch chronix2grid with T mode. Chronix2grid stopped before dispatch computation. You should launch xithout letter T in mode"
MSG_NO_DISPATCH_BACKEND = "Dispatch Backend is set to None - no dispatch has been computed"

class GeneratorBackend:
    """
    Class that gathers the Backends of the different generation processes.
    It will allow to generate each step successively, thanks to its method :func:`GeneratorBackend.run`.
    It will load and check the parameters for each step thanks to instances of :class:`chronix2grid.config.ConfigManager`
    It will then do the proper generation thanks to its methods :func:`GeneratorBackend.do_l`, :func:`GeneratorBackend.do_r`,
    :func:`GeneratorBackend.do_d` and :func:`GeneratorBackend.do_t`. This methods rely on other specific backends.

    All the attributes are static variables passed via module :class:`chronix2grid.constants`.
    This is where you can choose all the type of objects that will be used by *Chronix2Grid*

    Attributes
    ----------
    general_config_manager: :class:`chronix2grid.config.ConfigManager`
        Class inheriting from ConfigManager that loads and checks the general parameters of the overall generation process, such as time resolution.
    load_config_manager: :class:`chronix2grid.config.ConfigManager`
        Class inheriting from ConfigManager that loads and checks the load generation parameters
    res_config_manager: :class:`chronix2grid.config.ConfigManager`
        Class inheriting from ConfigManager that loads and checks the renewable (solar and wind) production parameters
    loss_config_manager: :class:`chronix2grid.config.ConfigManager`
        Class inheriting from ConfigManager that loads and checks the power loss generation parameters
    dispatch_config_manager: :class:`chronix2grid.config.ConfigManager`
        Class inheriting from ConfigManager that loads and checks the dispatch parameters
    dispatcher_class: :class:`chronix2grid.generation.dispatch.EconomicDispatch.Dispatcher`
        Class inheriting from Dispatcher that provides an API to a DispatchBackend in order to compute a proper dispatch
    consumption_backend_class
        A class that embeds a load generation backend such as :class:`chronix2grid.generation.consumption.ConsumptionGeneratorBackend`
    renewable_backend_class
        A class that embeds a solar and wind generation backend such as :class:`chronix2grid.generation.renewable.RenewableBackend`
    loss_backend_class
        A class that embeds a power loss generation backend such as :class:`chronix2grid.generation.loss.LossBackend`
    dispatch_backend_class
        A class that embeds a dispatch backend such as :class:`chronix2grid.generation.dispatch.DispatchBackend`
    """
    def __init__(self):
        from chronix2grid import default_backend  # lazy import to avoid circular references
        self.general_config_manager = default_backend.GENERAL_CONFIG
        self.load_config_manager = default_backend.LOAD_GENERATION_CONFIG
        self.res_config_manager = default_backend.RENEWABLE_GENERATION_CONFIG
        self.loss_config_manager = default_backend.LOSS_GENERATION_CONFIG
        self.dispatch_config_manager = default_backend.DISPATCH_GENERATION_CONFIG

        self.dispatcher_class = default_backend.DISPATCHER

        self.consumption_backend_class = default_backend.LOAD_GENERATION_BACKEND
        self.dispatch_backend_class = default_backend.DISPATCH_GENERATION_BACKEND
        self.hydro_backend_class = default_backend.HYDRO_GENERATION_BACKEND
        self.renewable_backend_class = default_backend.RENEWABLE_GENERATION_BACKEND
        self.loss_backend_class = default_backend.LOSS_GENERATION_BACKEND

    # Call generation scripts n_scenario times with dedicated random seeds
    def run(self, case, n_scenarios, input_folder, output_folder, scen_names,
            time_params, mode='LRTK', scenario_id=None,
            seed_for_loads=None, seed_for_res=None, seed_for_disp=None):
        """
        Main function for chronics generation. It works with four steps: load generation (L), renewable generation (R, solar and wind), loss generation (D)
        and then dispatch computation (T) to get the whole energy mix. It writes the resulting chronics in the output_path in zipped csv format

        Parameters
        ----------
        case: ``str``
            name of case to study (must be a folder within input_folder)
        n_scenarios: ``int``
            number of desired scenarios to generate for the same timescale
        input_folder: ``str``
            path of folder containing inputs
        output_folder: ``str``
            path where outputs will be written (intermediate folder case/year/scenario will be used)
        mode: ``str``
            options to launch certain parts of the generation process : L load R renewable T thermal
        scenario_id: ``int`` or ``None``
            Id of scenario
        seed_for_loads: ``int`` or ``None``
            seed for the load generation process
        seed_for_res: ``int`` or ``None``
            seed for the renewable generation process
        seed_for_disp: ``int`` or ``None``
            seed for the dispatch generation process

        Returns
        -------
        params: ``dict``
            general parameters
        loads_charac: :class:`pandas.DataFrame`
            characteristics of consumption nodes in the grid used in generation
        prods_charac: :class:`pandas.DataFrame`
            characteristics of production nodes in the grid used in generation
        """

        utils.check_scenario(n_scenarios, scenario_id)

        print('=====================================================================================================================================')
        print('============================================== CHRONICS GENERATION ==================================================================')
        print('=====================================================================================================================================')

        # in multiprocessing, n_scenarios=1 here
        if n_scenarios >= 2:
            seeds_for_loads, seeds_for_res, seeds_for_disp = generation_utils.generate_seeds(
                n_scenarios, seed_for_loads, seed_for_res, seed_for_disp
            )
        else:
            seeds_for_loads = [seed_for_loads]
            seeds_for_res = [seed_for_res]
            seeds_for_disp = [seed_for_disp]

        # dispatch_input_folder, dispatch_input_folder_case, dispatch_output_folder = gu.make_generation_input_output_directories(input_folder, case, year, output_folder)
        general_config_manager = self.general_config_manager(
            name="Global Generation",
            root_directory=input_folder,
            input_directories=dict(case=case),
            required_input_files=dict(case=['params.json']),
            output_directory=output_folder
        )
        general_config_manager.validate_configuration()
        params = general_config_manager.read_configuration()

        params.update(time_params)
        params = generation_utils.updated_time_parameters_with_timestep(params, params['dt'])

        load_config_manager = self.load_config_manager(
            name="Loads Generation",
            root_directory=input_folder,
            input_directories=dict(case=case, patterns='patterns'),
            required_input_files=dict(case=['loads_charac.csv', 'params_load.json'],
                                      patterns=['load_weekly_pattern.csv']),
            output_directory=output_folder
        )
        load_config_manager.validate_configuration()
        params_load, loads_charac = load_config_manager.read_configuration()
        params_load.update(params)

        res_config_manager = self.res_config_manager(
            name="Renewables Generation",
            root_directory=input_folder,
            input_directories=dict(case=case, patterns='patterns'),
            required_input_files=dict(case=['prods_charac.csv', 'params_res.json'],
                                      patterns=['solar_pattern.npy']),
            output_directory=output_folder
        )
        params_res, prods_charac = res_config_manager.read_configuration()
        params_res.update(params)


        grid_folder = os.path.join(input_folder, case)

        loss = None

        ## Launch proper scenarios generation
        seeds_iterator = zip(seeds_for_loads, seeds_for_res, seeds_for_disp)

        for i, (seed_load, seed_res, seed_disp) in enumerate(seeds_iterator):

            if n_scenarios > 1:
                scenario_name = scen_names(i)
            else:
                scenario_name = scen_names(scenario_id)

            scenario_folder_path = os.path.join(output_folder, scenario_name)

            print("================ Generating " + scenario_name + " ================")
            if 'L' in mode:
                load, load_forecasted = self.do_l(scenario_folder_path, seed_load, params_load, loads_charac, load_config_manager)
                params.update(params_load)
            if 'R' in mode:
                prod_solar, prod_solar_forecasted, prod_wind, prod_wind_forecasted = self.do_r(scenario_folder_path, seed_res, params_res,
                                                                                               prods_charac,
                                                                                               res_config_manager)
                params.update(params_res)
            if 'D' in mode:
                loss_config_manager = self.loss_config_manager(
                    name="Loss",
                    root_directory=input_folder,
                    output_directory=output_folder,
                    input_directories=dict(params=case),
                    required_input_files=dict(params=['params_loss.json'])
                )

                self.do_d(input_folder, scenario_folder_path,
                                     load, prod_solar, prod_wind,
                                     params, loss_config_manager)
            if 'T' in mode:
                if self.dispatch_backend_class is None:
                    warnings.warn(MSG_NO_DISPATCH_BACKEND, UserWarning)
                else:
                    dispath_config_manager = self.dispatch_config_manager(
                        name="Dispatch",
                        root_directory=input_folder,
                        output_directory=output_folder,
                        input_directories=dict(params=case),
                        required_input_files=dict(params=['params_opf.json'])
                    )
                    dispath_config_manager.validate_configuration()
                    params_opf = dispath_config_manager.read_configuration()

                    dispatch_results = self.do_t(input_folder, scenario_name, load, prod_solar, prod_wind,
                                                 grid_folder, scenario_folder_path, seed_disp, params, params_opf, loss)

            print('\n')
        return params, loads_charac, prods_charac

    def do_l(self, scenario_folder_path, seed_load, params, loads_charac, load_config_manager):
        """
        Generates load chronics thanks to the backend in ``self.consumption_backend_class``

        Parameters
        ----------
        scenario_folder_path: ``str``
        seed_load :``int``
        params: ``dict``
        loads_charac: :class:`pandas.DataFrame`
        load_config_manager: :class:`chronix2grid.config.ConfigManager`

        Returns
        -------
        loads: :class:`pandas.DataFrame`
            generated loads chronics
        prods_charac: :class:`pandas.DataFrame`
            generated forecasted loads chronics (currently loads chronics with gaussian noise)
        """
        generator_loads = self.consumption_backend_class(scenario_folder_path, seed_load, params, loads_charac, load_config_manager,
                                                         write_results=True)
        load, load_forecasted = generator_loads.run()
        return load, load_forecasted

    def do_r(self, scenario_folder_path, seed_res, params, prods_charac, res_config_manager):
        """
        Generates load chronics thanks to the backend in ``self.renewable_backend_class``

        Parameters
        ----------
        scenario_folder_path: ``str``
        seed_res: ``int``
        params: ``dict``
        prods_charac: :class:`pandas.DataFrame`
        res_config_manager: :class:`chronix2grid.config.ConfigManager`

        Returns
        -------
        prod_solar: :class:`pandas.DataFrame`
            generated solar chronics
        prod_solar_forecasted: :class:`pandas.DataFrame`
            generated forecasted solar chronics
        prod_wind: :class:`pandas.DataFrame`
            generated wind chronics
        prod_wind_forecasted: :class:`pandas.DataFrame`
            generated forecasted wind chronics
        """
        generator_enr = self.renewable_backend_class(scenario_folder_path, seed_res, params,
                                                     prods_charac,
                                                     res_config_manager, write_results=True)

        prod_solar, prod_solar_forecasted, prod_wind, prod_wind_forecasted = generator_enr.run()
        return prod_solar, prod_solar_forecasted, prod_wind, prod_wind_forecasted

    def do_d(self, input_folder, scenario_folder_path,
                                     load, prod_solar, prod_wind,
                                     params, loss_config_manager):
        """
        Generates load chronics thanks to the backend in ``self.renewable_backend_class``

        Parameters
        ----------
        input_folder: ``str``
        scenario_folder_path: ``str``
        load: :class: `pandas.DataFrame`
        prod_solar: :class: `pandas.DataFrame`
        prod_wind: :class: `pandas.DataFrame`
        params: ``dict``
        loss_config_manager: :class:`chronix2grid.config.ConfigManager`

        Returns
        -------
        loss: :class:`pandas.DataFrame`
            generated loss chronics
        """

        generator_loss = self.loss_backend_class(input_folder, scenario_folder_path,
                                     load, prod_solar, prod_wind,
                                     params, loss_config_manager)
        loss = generator_loss.run()
        return loss

    def do_t(self, input_folder, scenario_name, load, prod_solar, prod_wind, grid_folder,
             scenario_folder_path, seed_disp, params, params_opf, loss):
        """
        Computes production chronics based on a dispatch computation. It uses a dispatcher object as an environment for simulation and
        ``self.dispatch_backend_class`` for computation

        Parameters
        ----------
        dispatcher: :class:`chronix2grid.dispatch.EconomicDispatch.Dispatcher`
        scenario_name: ``str``
        load: :class:`pandas.DataFrame`
        prod_solar: :class:`pandas.DataFrame`
        prod_wind: :class:`pandas.DataFrame`
        grid_folder: ``str``
        scenario_folder_path: ``str``
        seed_disp: ``int``
        params: ``dict``
        params_opf: ``dict``
        loss: :class:`pandas.DataFrame`

        Returns
        -------
        dispatch_results: :class:`collection.namedtuple`
            contains resulting production chronics and terminal conditions from the optimization engine
        """

        prods = pd.concat([prod_solar, prod_wind], axis=1)
        res_names = dict(wind=prod_wind.columns, solar=prod_solar.columns)
        grid_path = os.path.join(grid_folder, constants.GRID_FILENAME)
        # grid_path = grid_folder
        dispatcher = EconomicDispatch.init_dispatcher_from_config_dataframe(grid_path, input_folder,self.dispatcher_class, params_opf)
        dispatcher.chronix_scenario = EconomicDispatch.ChroniXScenario(load, prods, res_names,
                                                                       scenario_name, loss)

        generator_dispatch = self.dispatch_backend_class(dispatcher, scenario_folder_path,
                                                 grid_folder, seed_disp, params, params_opf)
        dispatch_results = generator_dispatch.run()
        return dispatch_results



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

"""Class for the economic dispatch framework. Allows to parametrize and run
an economic dispatch based on RES and consumption time series"""

import numpy as np
import pandas as pd
from collections import namedtuple
import pypsa

from ._EDispatch_L2RPN2020.run_economic_dispatch import main_run_disptach

## Dépendances à Chronix2Grid
from chronix2grid.generation.dispatch.EconomicDispatch import Dispatcher
from chronix2grid.generation.dispatch.utils import RampMode

DispatchResults = namedtuple('DispatchResults', ['chronix', 'terminal_conditions'])


class PypsaDispatcher(Dispatcher, pypsa.Network):
    """
    Inheriting from Dispatcher to implement abstract methods thanks to Pypsa API
    Wrapper around a pypsa.Network to add higher level methods
    """

    # PATCH
    # to avoid problems for respecting pmax and ramps when rounding production values in chronics at the end, we apply a correcting factor
    PmaxCorrectingFactor = 1
    RampCorrectingFactor = 0.1
        
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add('Bus', 'node')
        self.add('Load', name='agg_load', bus='node')
        self._env = None  # The grid2op environment when instanciated with from_gri2dop_env
        self._df = None
        self._chronix_scenario = None
        self._simplified_chronix_scenario = None
        self._has_results = False
        self._has_simplified_results = False
        
        self._pmax_solar = None
        self._pmax_wind = None

    @classmethod
    def from_gri2op_env(cls, grid2op_env):
        """
        Implements the abstract method of *Dispatcher*

        Parameters
        ----------
        grid2op_env

        Returns
        -------
        net: :class:`pypsa.Network`
        """
        net = cls()
        net._env = grid2op_env

        carrier_types_to_exclude = ['wind', 'solar']
        for i, generator in enumerate(grid2op_env.name_gen):
            gen_type = grid2op_env.gen_type[i]
            if gen_type not in carrier_types_to_exclude:
                p_max = grid2op_env.gen_pmax[i]
                pnom = p_max - cls.PmaxCorrectingFactor
                
                rampUp = (grid2op_env.gen_max_ramp_up[i] - cls.RampCorrectingFactor) / p_max
                rampDown = (grid2op_env.gen_max_ramp_down[i] - cls.RampCorrectingFactor) / p_max
                
                net.add(
                    class_name='Generator', name=generator, bus='node',
                    p_nom=pnom, carrier=grid2op_env.gen_type[i],
                    marginal_cost=grid2op_env.gen_cost_per_MW[i],
                    ramp_limit_up=rampUp,
                    ramp_limit_down=rampDown,
                )

        # add total wind and solar (for curtailment)
        net._pmax_solar = np.sum([grid2op_env.gen_pmax[i] 
                              for i in range(grid2op_env.n_gen)
                              if grid2op_env.gen_type[i] == "solar"])
        net.add('Generator',
                name='agg_solar',
                bus='node',
                carrier="solar",
                marginal_cost=0.,
                p_nom=net._pmax_solar
                )
        net._pmax_wind = np.sum([grid2op_env.gen_pmax[i] 
                              for i in range(grid2op_env.n_gen)
                              if grid2op_env.gen_type[i] == "wind"])
        net.add('Generator',
                name='agg_wind',
                bus='node',
                carrier="wind",
                marginal_cost=0.1,  # we prefer to curtail the wind if we have the choice
                # that's because solar should be distributed on the grid
                p_nom=net._pmax_wind
        )
        
        return net

    @classmethod
    def from_dataframe(cls, env_df):
        """
        Implements the abstract method of *Dispatcher*

        Parameters
        ----------
        grid2op_env

        Returns
        -------
        net: :class:`pypsa.Network`
        """
        net = cls()
        net._df = env_df

        carrier_types_to_exclude = ['wind', 'solar']

        for i, (generator, gen_type, p_max, ramp_up, ramp_down, gen_cost_per_MW) in enumerate(zip(env_df['name'],
                                            env_df['type'],
                                            env_df['pmax'],
                                            env_df['max_ramp_up'],
                                            env_df['max_ramp_down'],
                                            env_df['cost_per_mw'])):
            if gen_type not in carrier_types_to_exclude:
                pnom = (p_max - cls.PmaxCorrectingFactor)
                rampUp = (ramp_up- cls.RampCorrectingFactor) / p_max
                rampDown = (ramp_down - cls.RampCorrectingFactor) / p_max

                net.add(
                    class_name='Generator',
                    name=generator,
                    bus='node',
                    p_nom=pnom,
                    carrier=gen_type,
                    marginal_cost=gen_cost_per_MW,
                    ramp_limit_up=rampUp,
                    ramp_limit_down=rampDown,
                )

        # add total wind and solar (for curtailment)
        net._pmax_solar = np.sum([p_max
                              for i, (gen_type, p_max) in enumerate(zip(env_df['type'],
                                                                        env_df['pmax']))
                              if gen_type == "solar"])
        net.add('Generator',
                name='agg_solar',
                bus='node',
                carrier="solar",
                marginal_cost=0.,
                p_nom=net._pmax_solar
                )
        
        net._pmax_wind = np.sum([p_max
                              for i, (gen_type, p_max) in enumerate(zip(env_df['type'],
                                                                        env_df['pmax']))
                              if gen_type == "wind"])
        net.add('Generator',
                name='agg_wind',
                bus='node',
                carrier="wind",
                marginal_cost=0.1,  # we prefer to curtail the wind if we have the choice
                # that's because solar should be distributed on the grid
                p_nom=net._pmax_wind
        )
        return net


    def run(self,
            load, 
            total_solar,
            total_wind,
            params,
            gen_constraints=None,
            ramp_mode=RampMode.hard,
            by_carrier=False,
            gen_min_pu_t=None,
            gen_max_pu_t=None,
            **kwargs):
        """
        Implements the abstract method of *Dispatcher*

        Returns
        -------
        simplified_net: :class:`pypsa.Network`
        """
        if total_solar is not None:
            total_solar = total_solar / self._pmax_solar
        if total_wind is not None:
            total_wind = total_wind / self._pmax_wind
        prods_dispatch, terminal_conditions, marginal_prices = \
            main_run_disptach(
                self if not by_carrier else self.simplify_net(),
                load, total_solar, total_wind, 
                params, gen_constraints, ramp_mode,
                gen_min_pu_t=gen_min_pu_t, gen_max_pu_t=gen_max_pu_t,
                **kwargs)
        if prods_dispatch is None or marginal_prices is None:
            return None
        
        if by_carrier:
            self._simplified_chronix_scenario = self._chronix_scenario.simplify_chronix()
            self._simplified_chronix_scenario.prods_dispatch = prods_dispatch
            self._simplified_chronix_scenario.marginal_prices = marginal_prices
            results = self._simplified_chronix_scenario
            self._has_simplified_results = True
            self._has_results = False
        else:
            self._chronix_scenario.prods_dispatch = prods_dispatch
            self._chronix_scenario.marginal_prices = marginal_prices
            results = self._chronix_scenario
            self._has_results = True
            self._has_simplified_results = False
        if self._env is None:
            self.reset_ramps_from_dataframe()
        else:
            self.reset_ramps_from_grid2op_env()
        
        return DispatchResults(chronix=results, terminal_conditions=terminal_conditions)

    def simplify_net(self):
        """
        Implements the abstract method of *Dispatcher*
        """
        carriers = self.generators.carrier.unique()
        simplified_net = PypsaDispatcher()
        for carrier in carriers:
            names = self.generators[self.generators.carrier == carrier].index.tolist()

            gens = self.generators.loc[
                names,
                ['p_nom', 'ramp_limit_up', 'ramp_limit_down', 'marginal_cost']
            ]
            gens['ramp_up_mw'] = gens['p_nom'] * gens['ramp_limit_up']
            gens['ramp_down_mw'] = gens['p_nom'] * gens['ramp_limit_down']
            params = gens.agg(['sum', 'mean'])
            simplified_net.add(
                class_name='Generator', name=carrier, bus='node',
                p_nom=params.loc['sum', 'p_nom'], carrier=carrier,
                marginal_cost=params.loc['mean', 'marginal_cost'],
                ramp_limit_up=params.loc['sum', 'ramp_up_mw'] / params.loc['sum', 'p_nom'],
                ramp_limit_down=params.loc['sum', 'ramp_down_mw'] / params.loc['sum', 'p_nom'],
            )
        simplified_net._hydro_file_path = self._hydro_file_path
        simplified_net._min_hydro_pu = self._min_hydro_pu.iloc[:, 0]
        simplified_net._max_hydro_pu = self._max_hydro_pu.iloc[:, 0]

        print('simplified dispatch by carrier')
        full_ramp = simplified_net.generators['p_nom'] * simplified_net.generators['ramp_limit_up']
        df_full_ramp = pd.DataFrame({'full_ramp': full_ramp})

        print(pd.concat(
            [
                simplified_net.generators[['p_nom', 'ramp_limit_up',
                                           'ramp_limit_down', 'marginal_cost']],
                df_full_ramp
            ], axis=1))
        return simplified_net



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)


from chronix2grid.generation._dispatch._PypsaDispatchBackend._EDispatch_L2RPN2020 import main_run_disptach, run_opf, RampMode



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

"""
This file gathers constants related to directories and files that chronix2grid
reads/writes. It also defines the ConfigManager and the Backend used for each step of generation process (LRDT)

.. note::
    For the moment you can only change the backend for renewable generation and the associated config manager.

    Switch from RenewableBackend to RenewableBackendGAN and from ResConfigManager to ResConfigManagerGan

    Note that this would require tensorflow 1.15.4 which is an optional dependency of chronix2grid
"""

##############################################
################## NAMES #####################
##############################################

GENERATION_FOLDER_NAME = 'generation'
KPI_FOLDER_NAME = 'kpi'
KPI_IMAGES_FOLDER_NAME = 'images'

DEFAULT_OUTPUT_FOLDER_NAME = 'output'
DEFAULT_INPUT_FOLDER_NAME = 'getting_started/example/input'

SCENARIO_FOLDER_BASE_NAME = 'Scenario'

SEEDS_FILE_NAME = 'seeds_info.json'

FLOATING_POINT_PRECISION_FORMAT = '%.1f'

TIME_STEP_FILE_NAME = 'time_interval.info'

REFERENCE_ZONE = 'France'

GRID_FILENAME = 'grid.json'



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

##############################################
################ BACKENDS ####################
##############################################

from chronix2grid.generation.consumption.ConsumptionGeneratorBackend import ConsumptionGeneratorBackend
from chronix2grid.generation.renewable.RenewableBackend import RenewableBackend
from chronix2grid.generation.loss.LossBackend import LossBackend
from chronix2grid.generation.dispatch.DispatchBackend import DispatchBackend
from chronix2grid.config import GeneralConfigManager, LoadsConfigManager, ResConfigManager, LossConfigManager, DispatchConfigManager

GENERAL_CONFIG = GeneralConfigManager

#### LOAD (L) ####
LOAD_GENERATION_CONFIG = LoadsConfigManager
LOAD_GENERATION_BACKEND = ConsumptionGeneratorBackend

#### RENEWABLE - SOLAR AND WIND (R) ####
# from RenewableGANBackend.config import ResConfigManagerGan
RENEWABLE_GENERATION_CONFIG = ResConfigManager #ResConfigManagerGan #ResConfigManager
# from RenewableGANBackend.backend import RenewableBackendGAN
RENEWABLE_GENERATION_BACKEND = RenewableBackend #RenewableBackendGAN # RenewableBackend

#### LOSS (D) ####
LOSS_GENERATION_CONFIG = LossConfigManager
LOSS_GENERATION_BACKEND = LossBackend

#### DISPATCH - HYDRO, THERMAL, NUCLEAR (T) ####
DISPATCH_GENERATION_CONFIG = DispatchConfigManager
HYDRO_GENERATION_BACKEND = None

from chronix2grid.generation.dispatch.PypsaDispatchBackend import PypsaDispatcher
DISPATCHER = PypsaDispatcher
DISPATCH_GENERATION_BACKEND = DispatchBackend

#### KPI (K) ####
RENEWABLE_NINJA_REFERENCE_FOLDER = 'renewable_ninja'
GAN_TRAINING_SET_REFERENCE_FOLDER = 'GAN_training_data'



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
from enum import Enum
import copy

import numpy as np

class RampMode(Enum):
    """
    Encodes the level of complexity of the ramp constraints to apply for
    the economic dispatch
    """
    none = -1
    easy = 0
    medium = 1
    hard = 2


def make_scenario_input_output_directories(input_folder, output_folder, scenario_name):
    os.makedirs(os.path.join(input_folder, scenario_name), exist_ok=True)
    os.makedirs(os.path.join(output_folder, scenario_name), exist_ok=True)
    return os.path.join(input_folder, scenario_name), os.path.join(output_folder, scenario_name)

def modify_hydro_ramps(env_df, hydro_dividing_factor=1., decimals=1):
    rampup_hydro = env_df.loc[env_df['type']=='hydro','max_ramp_up'].values
    rampdown_hydro = env_df.loc[env_df['type'] == 'hydro', 'max_ramp_up'].values

    env_df.loc[env_df['type'] == 'hydro', 'max_ramp_up'] = np.round(rampup_hydro/hydro_dividing_factor, decimals)
    env_df.loc[env_df['type'] == 'hydro', 'max_ramp_down'] = np.round(rampdown_hydro / hydro_dividing_factor, decimals)

    return env_df

def modify_slack_characs(env_df, slack_name, p_max_reduction=0, ramp_reduction=0, decimals = 1):
    slack_pmax = env_df.loc[env_df['name'] == slack_name, 'pmax'].values[0]
    slack_rampup_max = env_df.loc[env_df['name'] == slack_name, 'max_ramp_up'].values[0]
    slack_rampdown_max = env_df.loc[env_df['name'] == slack_name, 'max_ramp_down'].values[0]

    env_df.loc[env_df['name'] == slack_name, 'pmax'] =  np.round(slack_pmax - p_max_reduction, decimals)
    env_df.loc[env_df['name'] == slack_name, 'max_ramp_up'] = np.round(slack_rampup_max - ramp_reduction, decimals)
    env_df.loc[env_df['name'] == slack_name, 'max_ramp_down'] = np.round(slack_rampdown_max - ramp_reduction,
                                                                                 decimals)
    return env_df

def add_noise_gen(prng, dispatch, gen_cap, noise_factor):
    """ Add noise to opf dispatch to have more
    realistic real-time data

    Parameters
    ----------
    dispatch : dataframe
        Opf PyPSA output
    gen_cap : dataframe
        Maximun capacity for gen
    noise_factor : float
        Noise factor applied to every gen col

    Returns
    -------
    dataframe
        Distpach with noise
    """
    dispatch_new = copy.deepcopy(dispatch)  # dispatch.copy(deep=True)

    # variance_per_col = gen_cap * noise_factor
    print('applying noise to forecast of ' + str(noise_factor) + ' %')
    for col in list(dispatch_new):
        # Check for values greater than zero
        # (means unit has been distpached)
        # only_dispatched_steps = dispatch_new[col][dispatch_new[col] > 0]
        # print(only_dispatched_steps)

        noise = prng.lognormal(mean=0.0, sigma=noise_factor, size=dispatch_new.shape[0])
        dispatch_new[col] = dispatch[col] * noise
    return dispatch_new.round(2)


# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
import warnings
import shutil
import numpy as np
import pandas as pd
import pathlib

import grid2op
from grid2op.Chronics import Multifolder, GridStateFromFileWithForecasts
from grid2op.Parameters import Parameters
from grid2op.Runner import Runner
from grid2op.Chronics import GridStateFromFile

import chronix2grid.constants as cst

def move_env_temporarily(scenario_output_folder, grid_path):

    scenario_name = pathlib.Path(scenario_output_folder).name
    grid_temporary_path = grid_path+'_'+scenario_name

    if os.path.exists(grid_temporary_path):
        warnings.warn("Had to delete a previous scenario grid temporary path in input data", UserWarning)
        shutil.rmtree(grid_temporary_path)
    #os.makedirs(grid_temporary_path, exist_ok=True)
    print("temporary copy of grid in "+str(grid_temporary_path))
    shutil.copytree(grid_path, grid_temporary_path)
    return grid_temporary_path


def move_chronics_temporarily(scenario_output_folder, grid_path):

    scenario_name = pathlib.Path(scenario_output_folder).name
    chronics_temporary_path = os.path.join(grid_path, 'chronics')

    os.makedirs(chronics_temporary_path, exist_ok=True)
    chronics_temporary_path = os.path.join(chronics_temporary_path, scenario_name)
    if os.path.exists(chronics_temporary_path):
        warnings.warn("Had to delete a previous chronic temporary path in input data", UserWarning)
        shutil.rmtree(chronics_temporary_path)
    print("temporary copy of chronics in "+str(chronics_temporary_path))
    shutil.copytree(scenario_output_folder, chronics_temporary_path)

def remove_temporary_chronics(grid_path):
    chronics_temporary_path = os.path.join(grid_path, 'chronics')
    shutil.rmtree(chronics_temporary_path)

def remove_simulation_data(scenario_folder_path):
    simulation_data_folder = os.path.join(scenario_folder_path, 'loss_simulation')
    shutil.rmtree(simulation_data_folder)

def create_or_replace_simulation_data_folder(scenario_folder_path):
    simulation_data_folder = os.path.join(scenario_folder_path, 'loss_simulation')
    if os.path.exists(simulation_data_folder):
        shutil.rmtree(simulation_data_folder)
    os.makedirs(simulation_data_folder, exist_ok=False)
    return simulation_data_folder

def search_chronic_num_from_name(scenario_name, env):
    found_id = None
    # Search scenario with provided name
    for id, sp in enumerate(env.chronics_handler.real_data.subpaths):
        sp_end = os.path.basename(sp)
        if sp_end == scenario_name:
            found_id = id
    return found_id

def run_grid2op_simulation_donothing(grid_path, agent_result_path,  nb_core = 1,write_results=False,agent_results_path=None):
    """

    :param grid_path (str): path to folder where grid.json and other information on grid are stored
    :param agent_result_path (str): path in which we want to write the result of Grid2op simulation (via serialized EpisodeData object)
    :param agent_type (str): type of grid2op agent to use in simulation. Can be reco for RecoPowerlineAgent or do-nothing for DoNothingAgent
    :param nb_core (int): Simulations can be paralelized on cores. By default, 1 core is used here because paralelization is made previously on chronic scenarios
    :return:
    """
    print('Start grid2op simulation to compute realistic loss on grid')

    # Récupération des paramètres
    NB_CORE = nb_core # TOUJOURS 1 car la parallélisation se fait sur les scénarii dans le main
    nb_episode = 1
    #simulation_data_folder = create_or_replace_simulation_data_folder(agent_result_path)

    print('Grid2op simulation for loss')

    try:
        from lightsim2grid.LightSimBackend import LightSimBackend
        backend = LightSimBackend()
    except:
        from grid2op.Backend import PandaPowerBackend
        backend = PandaPowerBackend()
        print("You might need to install the LightSimBackend (provisory name) to gain massive speed up")
    # don't disconnect powerline on overflow, the thermal limit are not set for now, it would not make sens
    param = Parameters()
    param.init_from_dict({"NO_OVERFLOW_DISCONNECTION": True})

    # Load env and apropriate chronix scenario
    env = grid2op.make(grid_path,
                       param=param, backend=backend, test=True,data_feeding_kwargs={"gridvalueClass": GridStateFromFile})
    scenario_name = pathlib.Path(agent_result_path).name
    scen_id = search_chronic_num_from_name(scenario_name, env)
    env.set_id(scen_id)
    #env.reset()

    #env = grid2op.make(grid_path = grid_path, chronics_path = scenario_output_folder,
     #                  param=param, backend=backend, test=True)
    # If you remove the "GridStateFromFileWithForecasts", from above, chronics will NOT be loaded properly.
    # GridStateFromFileWithForecasts is the format used for the competition, so it is mandatory that this works!
    # WITHOUT ANY MODIFICATIONS

    # Beside the environment should be able to load all data generated, and not one episode.
    # so please look in grid2op for compatible formats. This is not a valid format.

    ##########
    #MAKE SURE THAT no maintenance and no attacks will run here. Check the config.py in your grid2op env folder
    #if (agent_type == 'reco'):
    #    #runner = Runner(**env.get_params_for_runner(), agentClass=DoNothingAgent)
    #    runner = Runner(**env.get_params_for_runner(), agentClass=RecoPowerlineAgent)
    #else:
    runner = Runner(**env.get_params_for_runner())
    # do regular computation as you would with grid2op
    path_save = None
    if(write_results):
        path_save=os.path.join(agent_results_path,'agent_results')
        os.makedirs(path_save, exist_ok=True)

    name_chron, cum_reward, nb_time_step, episode_data =runner.run_one_episode(path_save=path_save,
                             indx=scen_id,
                             pbar=True,
                             detailed_output=True)
    #res = runner.run(nb_episode=nb_episode, nb_process=NB_CORE, pbar=True, add_detailed_output=True)
    #                 #path_save=simulation_data_folder
    #id_chron, name_chron, cum_reward, nb_timestep, max_ts, episode_data = res.pop()
    print('---- end of simulation')

    return episode_data

def correct_scenario_loss(scenario_folder_path, params_opf, grid_path, data_this_episode):
    print('Start realistic loss correction from simulation results')

    # Load simulation data
    #data_this_episode = EpisodeData.from_disk(os.path.join(agent_result_path, 'loss_simulation'), 'Scenario_0')
    slack_name = params_opf["nameSlack"]
    id_slack = params_opf["idxSlack"]

    # Get gen constraints
    observations = [obs for obs in data_this_episode.observations]
    if observations[0] is None: # Quick hack because a None appears in observations with grid2op 1.5.0 - don't have time to handle it
        observations[0] = observations[1]
    first_obs = observations[0]
    pmax = first_obs.gen_pmax[id_slack] #+ params_opf['pmax_margin']
    pmin = first_obs.gen_pmin[id_slack] #max(first_obs.gen_pmin[id_slack] - params_opf['pmin_margin'],0)
    ramp_up = first_obs.gen_max_ramp_up[id_slack] #+ params_opf['rampup_margin']
    ramp_down = first_obs.gen_max_ramp_down[id_slack] #+ params_opf['rampdown_margin']

    # Get corrected dispatch prod
    prods_p = pd.DataFrame(np.array([obs.prod_p for obs in observations]))
    prodSlack = prods_p[id_slack]

    # Get dispatch prods before runner in chronix
    OldProdsDf = pd.read_csv(os.path.join(scenario_folder_path, 'prod_p.csv.bz2'), sep=';')
    OldProdsForecastDf = pd.read_csv(os.path.join(scenario_folder_path, 'prod_p_forecasted.csv.bz2'), sep=';')

    ##correction term
    newProdsDf = OldProdsDf
    newProdsForecastDf = OldProdsForecastDf
    CorrectionLosses = prodSlack - OldProdsDf[slack_name]
    print('maximum compensation for slack before correction in MW: ' + str(CorrectionLosses.abs().max()))
    print('average compensation for slack before correction in MW: ' + str(CorrectionLosses.mean()))
    print('median compensation for slack before correction in MW: ' + str(CorrectionLosses.abs().median()))
    print('min compensation for slack before correction in MW: ' + str(CorrectionLosses.min()))

    # apply correction on slack bus generator
    newProdsDf[slack_name] = OldProdsDf[slack_name] + CorrectionLosses
    newProdsForecastDf[slack_name] = OldProdsForecastDf[slack_name] + CorrectionLosses

    # Log the correction
    CorrectionLosses_df = pd.DataFrame({'adjusted_loss_p':CorrectionLosses})
    CorrectionLosses_df.to_csv(os.path.join(scenario_folder_path, 'adjusted_loss.csv.bz2'), sep=';', index = False)

    # Check constraints
    violations_message, bool = check_slack_constraints(newProdsDf[slack_name], pmax, pmin, ramp_up, ramp_down)
    if bool:
        if params_opf['early_stopping_mode']:
            remove_temporary_chronics(grid_path)
            raise ValueError(violations_message)
        else:
            warnings.warn(violations_message, UserWarning)
            print("Warning - "+violations_message)


    # Serialization
    newProdsDf.to_csv(
            os.path.join(scenario_folder_path, "prod_p.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )

    newProdsForecastDf.to_csv(
        os.path.join(scenario_folder_path, "prod_p_forecasted.csv.bz2"),
        sep=';', index=False,
        float_format=cst.FLOATING_POINT_PRECISION_FORMAT
    )

    print('---- end of loss correction ')
    return newProdsDf, newProdsForecastDf

def check_slack_constraints(prod_p, pmax, pmin, ramp_up, ramp_down):
    msg = "Loss correction violates generator constraints: "
    bool = False

    # Pmax
    dep = max(prod_p-pmax)
    if dep > 0:
        bool = True
        msg += "Pmax + margin is violated with maximum of "+str(dep)+" MW - "

    # Pmin
    dep = min(prod_p-pmin)
    if dep < 0:
        bool = True
        msg += "Pmin - margin is violated with maximum of " + str(dep) + " MW - "

    # Ramp up
    ramps = prod_p.diff()
    ramps_up = ramps[ramps>0]
    dep = max(ramps_up - ramp_up)
    if dep > 0:
        bool = True
        msg += "Ramp up + margin is violated with maximum of " + str(dep) + " MW - "

    # Ramp down
    ramps_down = -1 * ramps[ramps < 0]
    dep = max(ramps_down - ramp_down)
    if dep > 0:
        bool = True
        msg += "Ramp down + margin is violated with maximum of " + str(dep) + " MW - "

    return msg, bool


# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import os
from setuptools import setup, find_packages

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    
    
setup(name='Chronix2Grid',
      version='1.1.0',
      description='A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)',
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
          "Intended Audience :: Developers",
          "Intended Audience :: Education",
          "Intended Audience :: Science/Research",
          "Natural Language :: English"
      ],
      keywords='ML powergrid optmization RL power-systems chronics generation production load network',
      author='Mario Jothy, Nicolas Megel, Vincent Renault, Benjamin Donnot',
      author_email=' mario.jothy@artelys.com',
      url="https://github.com/BDonnot/ChroniX2Grid",
      license='Mozilla Public License 2.0 (MPL 2.0)',
      packages=find_packages(),
      include_package_data=True,
      platforms=["Windows", "Linux", "Mac OS-X", "Unix"],
      install_requires=[
                        "click>=7.1.1",
                        "cufflinks>=0.17.3",
                        "decorator>=4.4.2",
                        "folium>=0.10.1",
                        "Grid2Op>=1.5.2",
                        "ipython-genutils>=0.2.0",
                        "ipywidgets>=7.5.1",
                        "json5>=0.9.3",
                        "jupyter>=1.0.0",
                        "matplotlib>=3.1.3",
                        "numpy>=1.18.3",
                        "pandas>=1.0.5",
                        "pandocfilters>=1.4.2",
                        "plotly>=4.5.2",
                        "Pyomo>=5.6.8",
                        "pytest>=6.2.2",
                        "pytest-tornasync>=0.6.0.post2",
                        "python-dateutil>=2.8.1",
                        "pytz>=2019.3",
                        "PyUtilib>=5.7.3",
                        "requests>=2.23.0",
                        "seaborn>=0.10.0",
                        "scipy>=1.4.1",
                        "widgetsnbextension>=3.5.1",
                        "lightsim2grid",
                        "pypsa",
                        "cvxpy"
                        ],
      extras_require = {
                        "optional": [
                            "ligthsim2grid"
                        ],
                        "docs": [
                            "numpydoc>=0.9.2",
                            "sphinx>=2.4.4",
                            "sphinx-rtd-theme>=0.4.3",
                            "sphinxcontrib-trio>=1.1.0",
                            "autodocsumm>=0.1.13",
                            "pypsa>=0.17.0"
                        ]
                    },
      zip_safe=False,
      package_data={'chronix2grid':['getting_started/example/input/generation/case118_l2rpn_neurips_1x/*',
                                    'getting_started/example/input/generation/patterns/*',
                                    'getting_started/example/input/kpi/case118_l2rpn_neurips_1x/paramsKPI.json',
                                    'getting_started/example/input/kpi/case118_l2rpn_neurips_1x/France/eco2mix/*.csv',
                                    'getting_started/example/input/kpi/case118_l2rpn_neurips_1x/France/renewable_ninja/*.csv']},
      entry_points={'console_scripts': ['chronix2grid=chronix2grid.main:generate_mp']}
)



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

class RenewableBackendGAN:
    """
    Backend that generates solar and wind production chronics with a model based on trained neural networks (Generative
    Adversarial Networks or GAN).
    It takes general parameters about generation and specific parameters about how the networks have been trained.
    It uses tensorflow checkpoints in order to load and replay network objects

    .. warning::
        You should install tensorflow 1.15.4 to run this model, which is an optional dependency of chronix2grid

    Attributes
    ----------
    out_path: ``str``
        path to output folder for generated chronics
    seed: ``int``
    params: ``dict``
        dictionnary with the model parameters. It needs to contain keys **"dt", "planned_std", "model_name", "batch_size",
        "n_gens", "n_timesteps", "n_events","dim_inputs", "mu", "sigma", "network_folder"**.
        It has normally been read by a :class:`chronix2grid.config.ResConfigManagerGan` instance
    prods_charac: :class:`pandas.DataFrame`
        data frame with characteristics on wind and solar generators/power plants of the model
    res_config_manager: :class:`chronix2grid.config.ResConfigManagerGan`
    write_results: ``bool``
    """
    def __init__(self, out_path, seed, params, prods_charac, res_config_manager, write_results):
        self.write_results = write_results
        self.prods_charac = prods_charac
        self.res_config_manager = res_config_manager
        self.params = params
        self.seed = seed
        self.out_path = out_path

    def run(self):
        from RenewableGANBackend.generate_solar_wind_gan import main_gan
        """
        Runs the generation model in ``chronix2grid.generation.renewable.generate_solar_wind_gan`` and writes chronics
        """
        return main_gan(self.out_path, self.seed, self.params, self.prods_charac, self.write_results)


# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)
import os
import numpy as np
import json
from multiprocessing import Pool

import grid2op
from chronix2grid.grid2op_utils.utils import generate_a_scenario, get_last_scenario_id
from numpy.random import default_rng

import pdb

def add_data(env: grid2op.Environment.Environment,
             seed=None,
             nb_scenario=1,
             nb_core=1,  # TODO
             with_loss=True):
    """This function adds some data to already existing scenarios.
    
    .. warning::
        You should not start this function twice. Before starting a new run, make sure the previous one has terminated (otherwise you might
        erase some previously generated scenario)

    Parameters
    ----------
    env : _type_
        The grid2op environment
    seed:
        The seed to use (the same seed is guaranteed to generate the same scenarios)
    nb_scenario: ``int``
        The number of scenarios to generate
    nb_core: ``int``
        The number of core you want to use (to speed up the generation process)
    with_loss: ``bool``
        Do you make sure that the generated data will not be modified too much when running with grid2op (default = True).
        Setting it to False will speed up (by quite a lot) the generation process, but will degrade the data quality.
        
    """
    # required parameters
    env_name = type(env).env_name
    output_dir = os.path.join(env.get_path_env(), "chronics")
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
        
    last_scen = get_last_scenario_id(output_dir)
    scen_ids = [f"{el}" for el in range(last_scen +1, last_scen + 1 + nb_scenario)]
    with open(os.path.join(env.get_path_env(), "scenario_params.json"), "r", encoding="utf-8") as f:
        dict_ref = json.load(f)
        
    dt = dict_ref["dt"]
    li_months = dict_ref["all_dates"]

    # generate the seeds
    if seed is not None:
        prng = default_rng(seed)
    else:
        prng = default_rng()
        
    load_seeds = []
    renew_seeds = []
    gen_p_forecast_seeds = []
    for scen_id in scen_ids:
        for start_date in li_months:
            load_seed, renew_seed, gen_p_forecast_seed = prng.integers(2**32 - 1, size=3)
            load_seeds.append(load_seed)
            renew_seeds.append(renew_seed)
            gen_p_forecast_seeds.append(gen_p_forecast_seed)
    
    # generate the data
    # TODO multi proc
    path_env = env.get_path_env()
    name_gen = env.name_gen
    gen_type = env.gen_type
    errors = {}
    argss = []
    for j, scen_id in enumerate(scen_ids):
        for i, start_date in enumerate(li_months):
            seed_num = i + j * len(li_months)
            argss.append((path_env,
                          name_gen,
                          gen_type,
                          output_dir,
                          start_date,
                          dt,
                          scen_id,
                          load_seeds[seed_num],
                          renew_seeds[seed_num],
                          gen_p_forecast_seeds[seed_num],
                          with_loss
                          ))
    if nb_core == 1:
        for args in argss:
            path_env, name_gen, gen_type, output_dir, start_date, dt, scen_id, load_seed, renew_seed, \
                gen_p_forecast_seed, handle_loss = args
            res_gen = generate_a_scenario(path_env, name_gen, gen_type, output_dir, start_date, dt, scen_id, load_seed, renew_seed, 
                                          gen_p_forecast_seed, handle_loss)
            error_, *_ = res_gen
            if error_ is not None:
                print("=============================")
                print(f"     Error for {start_date} {scen_id}        ")
                print(f"{error_}")
                print("=============================")
                errors[f'{start_date}_{scen_id}'] = f"{error_}"
                
                # load previous data
                path_json_error = os.path.join(output_dir, "errors.json")
                if os.path.exists(path_json_error):
                    with open(path_json_error, "r", encoding="utf-8") as f:
                        err_tmp = json.load(f)
                    for k in err_tmp:
                        errors[k] = err_tmp[k]
                    
                # write the log
                with open(path_json_error, "w", encoding="utf-8") as f:
                    json.dump(errors, fp=f)
    else:
        with Pool(nb_core) as p:
            p.starmap(generate_a_scenario, argss)



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

from .generate_loss import main


class LossBackend:
    """
    Backend that generates loss simply based on a provided yearly pattern.
    The API provides user the ability to use power consumption, wind and solar productions as input for more complex loss modeling

    Attributes
    ----------
    input_folder: ``str``
        base input folder to get parameters and patterns
    scenario_folder_path: ``str``
        path to output folder for generated chronics
    load: :class:`pandas.DataFrame` or ``dict``
        generated load chronics by L mode
    prod_solar: :class:`pandas.DataFrame` or ``dict``
        generated solar chronics by R mode
    prod_wind: :class:`pandas.DataFrame` or ``dict``
        generated wind chronics by R mode
    params: ``dict``
        dictionnary with the model parameters. It needs to contain keys "dt" and  "loss_pattern"
    loss_config_manager: :class:`chronix2grid.config.LossConfigManager`
        config manager used to load specific patterns used for the model (yearly explicit loss pattern)
    write_results: ``bool``
    """
    def __init__(self, input_folder, scenario_folder_path,
                                     load, prod_solar, prod_wind,
                                     params, loss_config_manager, write_results = True):
        self.write_results = write_results
        self.loss_config_manager = loss_config_manager
        self.input_folder = input_folder
        self.params = params
        self.scenario_folder_path = scenario_folder_path
        self.load = load
        self.prod_solar = prod_solar
        self.prod_wind = prod_wind

    def run(self):
        """
        Runs the loss generation model in ``chronix2grid.generation.loss.generate_loss`` and writes loss chronics
        """
        self.loss_config_manager.validate_configuration()
        params_loss = self.loss_config_manager.read_configuration()
        return main(self.input_folder, self.scenario_folder_path,
                             self.load, self.prod_solar, self.prod_wind,
                             self.params, params_loss, write_results=self.write_results)



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Chronix2Grid, A python package to generate "en-masse" chronics for loads and productions (thermal, renewable)

import json
import os

import pandas as pd

from chronix2grid.config import ConfigManager


class ResConfigManagerGan(ConfigManager):
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        super(ResConfigManagerGan, self).__init__(name, root_directory, input_directories,
                                               output_directory, required_input_files)

    def read_configuration(self):
        """
        Reads parameters for :class:`chronix2grid.generation.renewable.RenewableBackendGAN`

            * *model_name_wind* - name of tensorflow serialized model for wind
            * *batch_size_wind*, *n_gens_wind*, *n_timesteps_wind* - dimensions of learning batches of wind networks
            * *n_events_wind* -  number of labels handled as input, representing arbitrary events in learning of wind
            * *dim_input_wind*, *mu_wind*, *sigma_wind* - characteristics of inpus (size of gaussian noise vector, mean and standard deviation)

        ALl the same parameters are expected for solar generation (with suffix *_solar*), that will use the same network architecture but with potentialy différent layout parameters

        Returns
        -------
        params_res: ``dict``
            dictionary of parameters
        prods_charac: :class:`pandas.DataFrame`
        """
        params_file_path = os.path.join(
            self.root_directory,
            self.input_directories['case'], 'params.json')
        network_folder = os.path.join(self.root_directory,
                                      self.input_directories['case'],
                                      'neural_network')
        params_file_path_gan = os.path.join(network_folder, 'paramsGAN.json')

        # Check timestep
        with open(params_file_path, 'r') as params_json:
            params_general = json.load(params_json)
        if params_general["dt"] % 60 != 0:
            raise ValueError('timesteps different from 60min or multiples of 60min are not supported yet with GAN. Please use another model')

        with open(params_file_path_gan, 'r') as params_json:
            params = json.load(params_json)

        for key, value in params.items():
            if key in ["mu_wind","sigma_wind","mu_solar","sigma_solar"]:
                params[key] = float(value)
            elif "model_name" not in key:
                try:
                    params[key] = int(value)
                except ValueError:
                    params[key] = pd.to_datetime(value, format='%Y-%m-%d')

        try:
            prods_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'prods_charac.csv'),
                sep=',')
            names = prods_charac['name']  # to generate error if separator is wrong

        except:
            prods_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'prods_charac.csv'),
                sep=';')
        params["network_folder"] = network_folder

        return params, prods_charac
